package com.restlet.client.async.impl;

import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.PromiseHandler;
import com.restlet.client.async.Promises;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.utils.Holder;
import com.restlet.client.utils.NotNull;
import com.restlet.client.utils.Nullable;

import static com.restlet.client.async.impl.DeferredImpl.LOG_SERVICE;

/**
 * Promises implementation,compatible with Promise/A+ proposal
 *
 * @param <C> type of value expected to be returned from the promise
 */
public class PromiseImpl<C> implements Promise<C> {

    private final Deferred<C> deferred;

    PromiseImpl(final Deferred<C> deferred) {
        this.deferred = deferred;
    }

    protected static void link(final Promise from, final Deferred to) {
        if (from == null) {
            to.resolve(null);
            return;
        }
        if (to.getState() != Deferred.State.PENDING) {
            return;
        }
        if (!from.isDone()) {
            from.getDeferred().addDeferredResolvedHandler(new com.restlet.client.async.Deferred.DeferredResolvedHandler() {
                @Override
                public void onResolve(final com.restlet.client.async.Deferred resolvedDeferred) {
                    if (to.getState() != Deferred.State.PENDING) {
                        return;
                    }
                    if (resolvedDeferred.getState() == Deferred.State.ERROR) {
                        to.reject(resolvedDeferred.getError());
                    } else if (resolvedDeferred.getState() == Deferred.State.SUCCESS) {
                        to.resolve(resolvedDeferred.getResult());
                    }
                }
            });
        } else {
            final com.restlet.client.async.Deferred deferred1 = from.getDeferred();
            if (deferred1.getState() == Deferred.State.ERROR) {
                to.reject(deferred1.getError());
            } else if (deferred1.getState() == Deferred.State.SUCCESS) {
                to.resolve(deferred1.getResult());
            }
        }

    }

    @Override
    public <T> Promise<T> then(@NotNull final PromiseHandler successHandler) {
        return then(successHandler, null);
    }

    @Override
    public <T> Promise<T> then(@Nullable final PromiseHandler successHandler,
                               @Nullable final PromiseHandler<Throwable> errorHandler) {
        if (successHandler == null && errorHandler == null) {
            throw new IllegalArgumentException("At least one of the handlers should be not null");
        }
        final Deferred futureDeferred = new DeferredImpl();
        if (isDone()) {
            notifyHandlers(futureDeferred, successHandler, errorHandler);
        } else {
            deferred.addDeferredResolvedHandler(new com.restlet.client.async.Deferred.DeferredResolvedHandler() {
                @Override
                public void onResolve(final com.restlet.client.async.Deferred deferred) {
                    notifyHandlers(futureDeferred, successHandler, errorHandler);
                }
            });
        }

        return futureDeferred.promise();
    }

    private void notifyHandlers(final Deferred futureDeferred, final PromiseHandler successHandler, final PromiseHandler<Throwable> errorHandler) {
        if (deferred.getState() == Deferred.State.SUCCESS) {
            if (successHandler != null) {
                try {
                    final Object handlerPromise = successHandler.on(deferred.getResult());
                    link(Promises.when(handlerPromise), futureDeferred);
                } catch (final Throwable e) {
                    futureDeferred.reject(e);
                }
            } else {
                futureDeferred.resolve(deferred.getResult());
            }
        } else if (deferred.getState() == Deferred.State.ERROR) {
            if (errorHandler != null) {
                try {
                    final Object handlerPromise = errorHandler.on(deferred.getError());
                    link(Promises.when(handlerPromise), futureDeferred);
                } catch (final Throwable e) {
                    futureDeferred.reject(e);
                }
            } else {
                futureDeferred.reject(deferred.getError());
            }
        }
    }

    @Override
    @Deprecated
    public <T> Promise<T> onError(@NotNull final PromiseHandler<Throwable> errorHandler) {
        return then(null, errorHandler);
    }

    @Override
    public Promise<C> recoverOnReject(final PromiseHandler<Throwable> errorHandler) {
        return then(null, errorHandler);
    }

    @Override
    public Promise<C> recoverOnReject(final Promise<C> promise) {
        return recoverOnReject(new PromiseHandler<Throwable>() {
            @Override
            public Object on(@Nullable final Throwable error) {
                return promise;
            }
        });
    }

    @Override
    public Promise<C> recoverOnReject(final C value) {
        return recoverOnReject(new PromiseHandler<Throwable>() {
            @Override
            public Object on(@Nullable final Throwable error) {
                return value;
            }
        });
    }

    @Override
    public Promise<C> recoverOnReject() {
        return recoverOnReject(new PromiseHandler<Throwable>() {
            @Override
            public Object on(@Nullable final Throwable error) {
                return null;
            }
        });
    }

    @Override
    public <T> Promise<T> flatRecoverOnReject(final Function<Throwable, Promise<T>> handler) {
        return then(null, new PromiseHandler<Throwable>() {
            @Override
            public Object on(@Nullable final Throwable error) {
                return handler.apply(error);
            }
        });
    }

    @Override
    public <T> Promise<T> cast() {
        return (Promise) this;
    }

    @Override
    public boolean isDone() {
        return deferred.getState() != Deferred.State.PENDING;
    }

    @Override
    public <T> Promise<T> flatMap(final Function<C, Promise<T>> handler) {
        if (isDone()) {
            if (deferred.getState() == Deferred.State.SUCCESS) {
                try {
                    final Promise<T> result = handler.apply(deferred.getResult());
                    if (result == null) {
                        return Promises.of();
                    }
                    return result;
                } catch (final Throwable e) {
                    return Promises.reject(e);
                }
            }
            return Promises.reject(deferred.getError());
        }
        final Deferred<T> futureDeferred = new DeferredImpl<>();
        this.deferred.addDeferredResolvedHandler(new Deferred.DeferredResolvedHandler() {
            @Override
            public void onResolve(final Deferred<?> deferred) {
                final Deferred.State state = deferred.getState();
                if (state == Deferred.State.SUCCESS) {
                    try {
                        link(handler.apply((C) deferred.getResult()), futureDeferred);
                    } catch (final Throwable e) {
                        futureDeferred.reject(e);
                    }
                } else {
                    futureDeferred.reject(deferred.getError());
                }
            }
        });
        return futureDeferred.promise();
    }

    @Override
    public <T> Promise<T> flatMapIfNotNull(final Function<C, Promise<T>> handler) {
        return flatMap(new Function<C, Promise<T>>() {
            @Override
            public Promise<T> apply(final C value) {
                if (value == null) {
                    return Promises.of();
                }

                return handler.apply(value);
            }
        });
    }

    @Override
    public <T> Promise<T> flatMap(final Promise<T> promise) {
        return flatMap(new Function<C, Promise<T>>() {
            @Override
            public Promise<T> apply(final C input) {
                return promise;
            }
        });
    }

    @Override
    public <T> Promise<T> map(final Function<C, T> handler) {
        return flatMap(new Function<C, Promise<T>>() {
            @Override
            public Promise<T> apply(final C input) {
                return Promises.of(handler.apply(input));
            }
        });
    }

    @Override
    public <T> Promise<T> mapIfNotNull(final Function<C, T> handler) {
        return flatMap(new Function<C, Promise<T>>() {
            @Override
            public Promise<T> apply(final C input) {
                if (input == null) {
                    return Promises.of();
                }
                return Promises.of(handler.apply(input));
            }
        });
    }

    @Override
    public Promise<C> doOnReject(final Consumer<Throwable> consumer) {
        try {
            if (getDeferred().getState() == Deferred.State.ERROR) {
                consumer.consume(getDeferred().getError());
            } else if (!isDone()) {
                deferred.addDeferredResolvedHandler(new Deferred.DeferredResolvedHandler() {
                    @Override
                    public void onResolve(final Deferred<?> deferred) {
                        doOnReject(consumer);
                    }
                });
            }
        } catch (final Throwable e) {
            LOG_SERVICE.error("Consumer threw an error when being notified of error", e);
        }
        return this;
    }

    @Override
    public Promise<C> doOnResolve(final Consumer<C> consumer) {
        return map(new Function<C, C>() {
            @Override
            public C apply(final C input) {
                consumer.consume(input);
                return input;
            }
        });
    }

    @Override
    public Promise<C> doOnResolveIfNotNull(final Consumer<C> consumer) {
        return doOnResolve(new Consumer<C>() {
            @Override
            public void consume(final C value) {
                if (value == null) {
                    return;
                }

                consumer.consume(value);
            }
        });
    }

    @Override
    public Promise<C> doFinally(final Consumer<Deferred> consumer) {
        try {
            if (isDone()) {
                consumer.consume(deferred);
            } else {
                deferred.addDeferredResolvedHandler(new Deferred.DeferredResolvedHandler() {
                    @Override
                    public void onResolve(final Deferred<?> deferred) {
                        doFinally(consumer);
                    }
                });
            }
        } catch (final Throwable e) {
            LOG_SERVICE.error("Consumer threw an error,when being notified of final state of promise", e);
        }
        return this;
    }

    @Override
    public Promise<Void> castToVoidPromise() {
        return map(new Function<C, Void>() {
            @Override
            public Void apply(final C whatever) {
                return null;
            }
        });
    }

    @Override
    public Promise<C> assignTo(final Holder<C> parentValueHolder) {
        return doOnResolve(new Consumer<C>() {
            @Override
            public void consume(final C value) {
                parentValueHolder.value = value;
            }
        });
    }

    @Override
    public com.restlet.client.async.Deferred getDeferred() {
        return deferred;
    }

}
