package com.restlet.client.script.impl;

import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.dao.EnvironmentsDao;
import com.restlet.client.dao.RepositoryDao;
import com.restlet.client.function.BiFunction;
import com.restlet.client.function.Function;
import com.restlet.client.function.Predicate;
import com.restlet.client.html.form.FormItemTo;
import com.restlet.client.html.form.FormItemType;
import com.restlet.client.model.EntityTo;
import com.restlet.client.net.http.request.HttpMethodDefinition;
import com.restlet.client.net.http.request.HttpRequestTo;
import com.restlet.client.net.http.request.impl.EffectiveHttpRequestTo;
import com.restlet.client.net.request.RequestHeaderTo;
import com.restlet.client.net.uri.UriQueryParamTo;
import com.restlet.client.net.uri.UriSchemeTo;
import com.restlet.client.net.uri.UriTo;
import com.restlet.client.net.uri.UriUtils;
import com.restlet.client.net.uri.impl.UriSchemeToImpl;
import com.restlet.client.script.ScriptService;
import com.restlet.client.utils.Maybe;
import com.restlet.client.utils.Objects;
import com.restlet.client.utils.Sequence;
import com.restlet.client.utils.StringUtils;

import java.util.ArrayList;
import java.util.List;

import static com.restlet.client.async.Promises.of;
import static com.restlet.client.net.uri.impl.UriSchemeToImpl.isSupportedScheme;
import static com.restlet.client.utils.EntityCloner.ENTITY_CLONER;
import static com.restlet.client.utils.FunctionalUtils.ifNotNull;
import static com.restlet.client.utils.Objects.isNullOrEmpty;
import static com.restlet.client.utils.StringUtils.isBlank;

public abstract class BaseRequestEvaluator {

    private final EnvironmentsDao environmentsDao;
    private final ScriptService scriptService;

    private final Function<String, String> encoder = new Function<String, String>() {
        @Override
        public String apply(final String input) {
            return encodeQueryString(input);
        }
    };

    protected BaseRequestEvaluator(final EnvironmentsDao environmentsDao, final ScriptService scriptService) {
        this.environmentsDao = environmentsDao;
        this.scriptService = scriptService;
    }

    protected Promise<String> evaluate(final String expression, final RepositoryDao repositoryDao) {
        return scriptService.evaluate(expression, environmentsDao, repositoryDao);
    }

    protected Promise<String> evaluate(final String value, final boolean isEvaluated, final RepositoryDao repositoryDao) {
        return isEvaluated ? evaluate(value, repositoryDao) : of(value);
    }

    protected abstract String decodeQueryString(String query);

    protected abstract String encodeQueryString(String query);

    public Promise<Boolean> isDynamic(final EntityTo entity, final RepositoryDao repositoryDao) {
        if (entity.getType() == EntityTo.Type.Request) {
            return isDynamic((HttpRequestTo) entity, repositoryDao);
        } else {
            return of(entity.getType() == EntityTo.Type.Scenario);
        }
    }

    public Promise<Boolean> isDynamic(final HttpRequestTo requestTo, final RepositoryDao repositoryDao) {
        if (requestTo == null) {
            return of(false);
        }

        return effective(requestTo, repositoryDao)
                .map(new Function<EffectiveHttpRequestTo, Boolean>() {
                    @Override
                    public Boolean apply(final EffectiveHttpRequestTo effectiveHttpRequestTo) {
                        return effectiveHttpRequestTo.isDynamic();
                    }
                });
    }

    private boolean isLocal(final HttpRequestTo request) {
        return request != null && UriUtils.isLoopbackAddress(request.getUri());
    }

    public Promise<EffectiveHttpRequestTo> effective(final HttpRequestTo requestTo, final RepositoryDao repositoryDao) {

        if (requestTo == null) {
            return of();
        }
        final HttpRequestTo effective = ENTITY_CLONER.clone(requestTo);
        return evaluate(requestTo, effective, repositoryDao)
                .map(new Function<Integer, EffectiveHttpRequestTo>() {
                    @Override
                    public EffectiveHttpRequestTo apply(final Integer nbEvaluatedExpressions) {
                        return new EffectiveHttpRequestTo(effective, nbEvaluatedExpressions > 0, isLocal(effective));
                    }
                });
    }

    private Promise<Integer> evaluate(final HttpRequestTo requestTo, final HttpRequestTo effective, final RepositoryDao repositoryDao) {
        final List<Promise<Integer>> nbDynamicExpressionsPromises = new ArrayList<>();

        // URI
        nbDynamicExpressionsPromises.add(evaluateUri(requestTo, effective.getUri(), repositoryDao));

        // method
        nbDynamicExpressionsPromises.add(evaluateMethod(requestTo, effective.getMethod(), repositoryDao));

        // headers
        nbDynamicExpressionsPromises.add(evaluateHeaders(effective.getHeaders(), repositoryDao));

        // body
        nbDynamicExpressionsPromises.add(evaluateBody(requestTo, effective, repositoryDao));

        return Promises.all(nbDynamicExpressionsPromises)
                .map(new Function<List<Integer>, Integer>() {
                    @Override
                    public Integer apply(final List<Integer> input) {
                        return Sequence.of(input).reduce(new BiFunction<Integer, Integer, Integer>() {
                            @Override
                            public Integer apply(final Integer nbTotalDynamicExpressions, final Integer nbDynamicExpressions) {
                                return nbTotalDynamicExpressions + nbDynamicExpressions;
                            }
                        }, 0);
                    }
                });
    }

    private Promise<Integer> evaluateMethod(final HttpRequestTo requestTo, final HttpMethodDefinition method, final RepositoryDao repositoryDao) {
        if (method == null || !method.isCustom()) {
            return of(0);
        }

        return evaluate(method.getName(), repositoryDao).map(new Function<String, Integer>() {
            @Override
            public Integer apply(final String evaluatedExpression) {
                method.setName(evaluatedExpression != null ? evaluatedExpression.toUpperCase() : null);
                return Objects.equals(requestTo.getMethod().getName(), evaluatedExpression) ? 0 : 1;
            }
        });
    }

    private Promise<Integer> evaluateBody(final HttpRequestTo requestTo, final HttpRequestTo effective, final RepositoryDao repositoryDao) {
        if (effective.getBody() == null || effective.getBody().getBodyType() == null) {
            return of(0);
        }

        final List<Promise<Boolean>> dynamicExpressionsPromises = new ArrayList<>();

        switch (effective.getBody().getBodyType()) {
            case File:
                // clone file body
                effective.getBody().setFileBody(requestTo.getBody().getFileBody());
                break;

            case Form:
                if (effective.getBody().getFormBody() != null && effective.getBody().getFormBody().getItems() != null && !effective.getBody().getFormBody().getItems().isEmpty()) {

                    int it = 0;

                    for (final FormItemTo item : effective.getBody().getFormBody().getItems()) {
                        if (item == null || !item.isEnabled()) {
                            it += 1;
                            continue;
                        }
                        if (!StringUtils.isBlank(item.getName())) {
                            final String itemName = item.getName();
                            dynamicExpressionsPromises.add(evaluate(item.getName(), !effective.isBodyEvaluationDisabled(), repositoryDao).map(new Function<String, Boolean>() {
                                @Override
                                public Boolean apply(final String evaluatedExpression) {
                                    item.setName(evaluatedExpression);
                                    return !effective.isBodyEvaluationDisabled() && !Objects.equals(itemName, evaluatedExpression);
                                }
                            }));
                        }

                        if (FormItemType.Text.equals(item.getType()) && !StringUtils.isBlank(item.getValue())) {
                            final String itemValue = item.getValue();
                            dynamicExpressionsPromises.add(evaluate(item.getValue(), !effective.isBodyEvaluationDisabled(), repositoryDao).map(new Function<String, Boolean>() {
                                @Override
                                public Boolean apply(final String evaluatedExpression) {
                                    item.setValue(evaluatedExpression);
                                    return !effective.isBodyEvaluationDisabled() && !Objects.equals(itemValue, evaluatedExpression);
                                }
                            }));
                        } else if (FormItemType.File.equals(item.getType())) {
                            // clone file body
                            item.setBlobValue(requestTo.getBody().getFormBody().getItems().get(it).getBlobValue());
                        }
                        it += 1;
                    }

                }
                break;

            case Text:
                if (!StringUtils.isBlank(effective.getBody().getTextBody())) {
                    dynamicExpressionsPromises.add(evaluate(effective.getBody().getTextBody(), !effective.isBodyEvaluationDisabled(), repositoryDao).map(new Function<String, Boolean>() {
                        @Override
                        public Boolean apply(final String evaluatedExpression) {
                            effective.getBody().setTextBody(evaluatedExpression);
                            return !effective.isBodyEvaluationDisabled() && !Objects.equals(requestTo.getBody().getTextBody(), evaluatedExpression);
                        }
                    }));
                }
                break;
        }

        return getNumberOfDynamicExpressions(dynamicExpressionsPromises);
    }

    private Promise<Integer> evaluateHeaders(final List<RequestHeaderTo> headers, final RepositoryDao repositoryDao) {
        if (Objects.isNullOrEmpty(headers)) {
            return of(0);
        }

        final List<Promise<Boolean>> dynamicExpressionsPromises = new ArrayList<>();

        for (final RequestHeaderTo header : headers) {
            if (header == null || !header.isEnabled()) {
                continue;
            }

            if (!StringUtils.isBlank(header.getName())) {
                final String headerName = header.getName();
                dynamicExpressionsPromises.add(evaluate(header.getName(), repositoryDao).map(new Function<String, Boolean>() {
                    @Override
                    public Boolean apply(final String evaluatedExpression) {
                        header.setName(evaluatedExpression);
                        return !Objects.equals(headerName, evaluatedExpression);
                    }
                }));
            }
            if (!StringUtils.isBlank(header.getValue())) {
                final String headerValue = header.getValue();
                dynamicExpressionsPromises.add(evaluate(header.getValue(), repositoryDao).map(new Function<String, Boolean>() {
                    @Override
                    public Boolean apply(final String evaluatedExpression) {
                        header.setValue(evaluatedExpression);
                        return !Objects.equals(headerValue, evaluatedExpression);
                    }
                }));
            }
        }

        return getNumberOfDynamicExpressions(dynamicExpressionsPromises);
    }

    private Promise<Integer> evaluateUri(final HttpRequestTo requestTo, final UriTo requestUri, final RepositoryDao repositoryDao) {
        if (requestUri == null) {
            return of(0);
        }

        final List<Promise<Boolean>> dynamicExpressionsPromises = new ArrayList<>();

        final UriSchemeTo originalScheme = ifNotNull(requestTo.getUri(),
                new Function<UriTo, UriSchemeTo>() {
                    @Override
                    public UriSchemeTo apply(final UriTo uri) {
                        return uri.getScheme();
                    }
                });

        requestUri.setScheme(originalScheme);

        if (originalScheme != null && !isSupportedScheme(originalScheme)) {
            dynamicExpressionsPromises.add(evaluate(originalScheme.getName(), repositoryDao)
                    .map(new Function<String, Boolean>() {
                        @Override
                        public Boolean apply(final String evaluatedScheme) {
                            if (isSupportedScheme(evaluatedScheme)) {
                                requestUri.setScheme(UriSchemeToImpl.valueOf(evaluatedScheme));
                                return true;
                            } else {
                                requestUri.setScheme(null);
                                return !Objects.equals(evaluatedScheme, originalScheme.getName());
                            }
                        }
                    }));
        }

        // URI host
        if (!isBlank(requestUri.getHost())) {
            dynamicExpressionsPromises.add(evaluate(requestUri.getHost(), repositoryDao).map(new Function<String, Boolean>() {
                @Override
                public Boolean apply(final String evaluatedExpression) {
                    requestUri.setHost(evaluatedExpression);
                    return !Objects.equals(requestTo.getUri().getHost(), evaluatedExpression);
                }
            }));
        }

        // URI path
        if (!isBlank(requestUri.getPath())) {

            dynamicExpressionsPromises.add(evaluate(requestUri.getPath(), repositoryDao).map(new Function<String, Boolean>() {
                @Override
                public Boolean apply(final String evaluatedExpression) {
                    requestUri.setPath(evaluatedExpression);
                    return !Objects.equals(requestTo.getUri().getPath(), evaluatedExpression);
                }
            }));
        }

        // URI fragment
        if (!isBlank(requestUri.getFragment())) {
            dynamicExpressionsPromises.add(evaluate(requestUri.getFragment(), repositoryDao).map(new Function<String, Boolean>() {
                @Override
                public Boolean apply(final String evaluatedExpression) {
                    requestUri.setFragment(evaluatedExpression);
                    return !Objects.equals(requestTo.getUri().getFragment(), evaluatedExpression);
                }
            }));
        }

        // URI query
        if (requestUri.getQuery() != null && !isNullOrEmpty(requestUri.getQuery().getItems())) {

            for (final UriQueryParamTo qp : requestUri.getQuery().getItems()) {

                if (qp == null || !qp.isEnabled()) {
                    continue;
                }

                if (!isBlank(qp.getName())) {

                    dynamicExpressionsPromises.add(Promises.combine(evaluate(qp.getName(), repositoryDao), evaluate(qp.getValue(), repositoryDao), new Promises.CombineCallback<String, String, Boolean>() {
                        @Override
                        public Boolean on(final String name, final String value) {
                            final boolean isDynamic = !Objects.equals(qp.getName(), name) || !Objects.equals(qp.getValue(), value);

                            if (qp.isEncoded()) {
                                qp.setName(name);
                                qp.setValue(value);
                            } else {
                                qp.setName(Maybe.ofThrowable(encoder, name).orElse(name));
                                qp.setValue(Maybe.ofThrowable(encoder, value).orElse(value));
                            }

                            return isDynamic;
                        }
                    }));
                }

            }
        }

        return getNumberOfDynamicExpressions(dynamicExpressionsPromises)
                .map(new Function<Integer, Integer>() {
                    @Override
                    public Integer apply(final Integer nbOfDynamicExpressions) {
                        UriUtils.ensureSchemeIsSet(requestUri);
                        return nbOfDynamicExpressions;
                    }
                });

    }

    private Promise<Integer> getNumberOfDynamicExpressions(final List<Promise<Boolean>> dynamicExpressionsPromises) {
        return Promises.all(dynamicExpressionsPromises)
                .map(new Function<List<Boolean>, Integer>() {
                    @Override
                    public Integer apply(final List<Boolean> dynamicExpressions) {
                        return Sequence.of(dynamicExpressions)
                                .filter(new Predicate<Boolean>() {
                                    @Override
                                    public boolean test(final Boolean isDynamic) {
                                        return isDynamic;
                                    }
                                })
                                .count();
                    }
                });
    }
}
