package com.restlet.client.tests.asserts.impl;

import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.dao.EnvironmentsDao;
import com.restlet.client.dao.RepositoryDao;
import com.restlet.client.function.BiFunction;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.html.HtmlKit;
import com.restlet.client.net.http.HttpClientResultTo;
import com.restlet.client.net.http.HttpDate;
import com.restlet.client.net.http.response.HttpResponseTo;
import com.restlet.client.platform.blob.Blob;
import com.restlet.client.platform.blob.BlobReader;
import com.restlet.client.platform.json.JsonEngine;
import com.restlet.client.platform.xml.XmlEngine;
import com.restlet.client.script.ScriptService;
import com.restlet.client.tests.asserts.AssertionEngine;
import com.restlet.client.tests.asserts.AssertionResultTo;
import com.restlet.client.tests.asserts.AssertionResultTo.State;
import com.restlet.client.utils.BlobUtils;
import com.restlet.client.utils.EntityUtils;
import com.restlet.client.utils.Maybe;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import static com.restlet.client.net.http.HeaderUtils.findHeaderValue;
import static com.restlet.client.utils.Objects.isNullOrEmpty;
import static com.restlet.client.utils.StringUtils.firstNotBlank;
import static com.restlet.client.utils.StringUtils.isBlank;
import static com.restlet.client.utils.StringUtils.joiner;

public class AssertionEngineImpl implements AssertionEngine {

    private final HttpDate httpDate;
    private final BlobReader blobReader;
    private final HtmlKit htmlKit;
    private final ScriptService scriptService;

    private final XmlAssertionEngineImpl xmlAssertionEngine;
    private final JsonAssertionEngineImpl jsonAssertionEngine;

    public AssertionEngineImpl(final JsonEngine jsonEngine, final XmlEngine xmlEngine, final HttpDate httpDate, final BlobReader blobReader, final HtmlKit htmlKit, final ScriptService scriptService) {
        super();

        this.httpDate = httpDate;
        this.blobReader = blobReader;
        this.htmlKit = htmlKit;
        this.scriptService = scriptService;
        this.xmlAssertionEngine = new XmlAssertionEngineImpl(xmlEngine, httpDate, blobReader, htmlKit);
        this.jsonAssertionEngine = new JsonAssertionEngineImpl(jsonEngine, httpDate, blobReader, htmlKit);
    }

    @Override
    public Promise<Void> evaluateAndCheck(final AssertionResultTo assertion, final HttpClientResultTo result,
                                          final EnvironmentsDao environmentsDao, final RepositoryDao repositoryDao) {
        assertion.setState(null);
        assertion.setEvaluatedPath(null);
        assertion.setEvaluatedValue(null);
        assertion.getMessages().clear();

        if (!EntityUtils.isEnabled(assertion) || assertion.getSubject() == null || assertion.getComparison() == null) {
            return Promises.of();
        }

        if (!isBlank(assertion.getPath())) {
            return scriptService.evaluate(assertion.getPath(), environmentsDao, repositoryDao)
                    .flatMap(new Function<String, Promise<Void>>() {
                        @Override
                        public Promise<Void> apply(final String evaluatedPath) {
                            assertion.setEvaluatedPath(evaluatedPath);
                            return evaluateAssertionValueAndCheck(assertion, result, environmentsDao, repositoryDao);
                        }
                    })
                    .flatRecoverOnReject(new Function<Throwable, Promise<Void>>() {
                        @Override
                        public Promise<Void> apply(final Throwable whatever) {
                            return evaluateAssertionValueAndCheck(assertion, result, environmentsDao, repositoryDao);
                        }
                    });
        }
        return evaluateAssertionValueAndCheck(assertion, result, environmentsDao, repositoryDao);
    }

    private Promise<Void> evaluateAssertionValueAndCheck(final AssertionResultTo assertion, final HttpClientResultTo result,
                                                         final EnvironmentsDao environmentsDao, final RepositoryDao repositoryDao) {

        if (!isBlank(assertion.getValue())) {
            return scriptService.evaluate(assertion.getValue(), environmentsDao, repositoryDao)
                    .flatMap(new Function<String, Promise<Void>>() {
                        @Override
                        public Promise<Void> apply(final String evaluatedValue) {
                            assertion.setEvaluatedValue(evaluatedValue);
                            return checkEvaluatedAssertion(assertion, result);
                        }
                    })
                    .flatRecoverOnReject(new Function<Throwable, Promise<Void>>() {
                        @Override
                        public Promise<Void> apply(final Throwable whatever) {
                            return checkEvaluatedAssertion(assertion, result);
                        }
                    });
        }
        return checkEvaluatedAssertion(assertion, result);
    }

    private Promise<Void> checkEvaluatedAssertion(final AssertionResultTo assertion, final HttpClientResultTo result) {
        if (assertion.getSubject() == null) {
            return Promises.of();
        }

        switch (assertion.getSubject()) {
            case Response:
                return checkResponse(assertion, result);

            case ResponseStatus:
                if (AssertionResultTo.ASSERTION_PATH_STATUS_MESSAGE.equals(assertion.getPath())) {
                    return checkStatusMessage(assertion, result != null ? result.getResponse() : null);
                }
                return checkStatusCode(assertion, result != null ? result.getResponse() : null);

            case ResponseBody:
                if (AssertionResultTo.ASSERTION_PATH_RESPONSE_BODY_LENGTH.equals(assertion.getPath())) {
                    return checkBodyLength(assertion, result != null ? result.getResponse() : null);
                }
                return checkBodyContent(assertion, result != null ? result.getResponse() : null);

            case ResponseJsonBody:
                return jsonAssertionEngine.checkBody(assertion, result != null ? result.getResponse() : null);

            case ResponseXmlBody:
                return xmlAssertionEngine.checkBody(assertion, result != null ? result.getResponse() : null);

            case ResponseHeader:
                return checkHeader(assertion, result != null ? result.getResponse() : null);

            default:
                break;
        }
        return Promises.of();
    }

    @Override
    public Promise<Void> evaluateAndCheck(final List<AssertionResultTo> assertions, final HttpClientResultTo result,
                                          final EnvironmentsDao environmentsDao, final RepositoryDao repositoryDao) {
        if (isNullOrEmpty(assertions)) {
            return Promises.of();
        }

        final ArrayList<Promise<Void>> promises = new ArrayList<>();

        for (final AssertionResultTo assertion : assertions) {
            promises.add(evaluateAndCheck(assertion, result, environmentsDao, repositoryDao));
        }

        return Promises.all(promises)
                .castToVoidPromise();
    }

    private Promise<Void> checkResponse(final AssertionResultTo assertion, final HttpClientResultTo response) {

        Integer expectedValue = null;

        final String stringValue = !isBlank(assertion.getEvaluatedValue()) ? assertion.getEvaluatedValue() : assertion.getValue();

        if (!isBlank(stringValue)) {
            try {

                expectedValue = Integer.valueOf(stringValue.trim());

            } catch (final Exception e) {
                assertion.setState(State.Error);
                assertion.addErrorMessage("Value &lt;" + htmlKit.escape(stringValue.trim()) + "&gt; is not a number.");
                return Promises.of();
            }
        }

        if (expectedValue == null) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("Value &lt;&gt; is not a number.");
            return Promises.of();
        }

        if (response == null) {
            return Promises.of();
        }

        if (response.getRequestSent() != null && response.getResponseReceived() != null) {

            final long latency = response.getResponseReceived().getTime() - response.getRequestSent().getTime();
            assertion.setActualValue(String.valueOf(latency));

            switch (assertion.getComparison()) {
                case Equals:
                    if (expectedValue == latency) {
                        assertion.setState(State.Ok);

                    } else {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Expected latency: &lt;" + expectedValue + "&gt; ms but was: &lt;" + latency + "&gt; ms.");
                    }

                    break;

                case DoesNotEqual:
                    if (expectedValue != latency) {
                        assertion.setState(State.Ok);

                    } else {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Expected latency: &lt;" + expectedValue + "&gt; ms is the same.");
                    }
                    break;

                case Less:
                    if (latency < expectedValue) {
                        assertion.setState(State.Ok);

                    } else {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Latency: &lt;" + latency + "&gt; ms is not less than: &lt;" + expectedValue + "&gt; ms.");
                    }
                    break;

                case LessOrEqual:
                    if (latency <= expectedValue) {
                        assertion.setState(State.Ok);

                    } else {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Latency: &lt;" + latency + "&gt; ms is not less or equal to: &lt;" + expectedValue + "&gt; ms.");
                    }
                    break;

                case Greater:
                    if (latency > expectedValue) {
                        assertion.setState(State.Ok);

                    } else {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Latency: &lt;" + latency + "&gt; ms is not greater than: &lt;" + expectedValue + "&gt; ms.");
                    }
                    break;

                case GreateOrEqual:
                    if (latency >= expectedValue) {
                        assertion.setState(State.Ok);

                    } else {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Latency: &lt;" + latency + "&gt; ms is not greater or equal to: &lt;" + expectedValue + "&gt; ms.");
                    }
                    break;

                default:
                    break;
            }
        }
        return Promises.of();
    }

    private Promise<Void> checkStatusMessage(final AssertionResultTo assertion, final HttpResponseTo response) {
        if (response == null) {
            return Promises.of();
        }

        final String expectedValue = !isBlank(assertion.getEvaluatedValue()) ? assertion.getEvaluatedValue() : assertion.getValue();

        final String statusMessage = response.getStatus() != null ? response.getStatus().getMessage() : null;
        assertion.setActualValue(statusMessage);

        switch (assertion.getComparison()) {
            case Equals:
                if ((!isBlank(expectedValue) && !isBlank(statusMessage) && expectedValue.equalsIgnoreCase(statusMessage.trim()))
                        || (isBlank(expectedValue) && isBlank(statusMessage))
                        ) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected response status message: &lt;" + (expectedValue != null ? htmlKit.escape(expectedValue) : "") + "&gt; but was: &lt;" + (statusMessage != null ? htmlKit.escape(statusMessage) : "") + "&gt;.");
                }
                break;

            case DoesNotEqual:
                if ((!isBlank(expectedValue) && !isBlank(statusMessage) && !expectedValue.equalsIgnoreCase(statusMessage.trim()))
                        || (isBlank(expectedValue) && !isBlank(statusMessage))
                        || (!isBlank(expectedValue) && isBlank(statusMessage))
                        ) {
                    assertion.setState(State.Ok);

                } else if (!isBlank(expectedValue)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status message: &lt;" + (htmlKit.escape(expectedValue)) + "&gt; is the same.");
                }
                break;

            case Exists:
                if (!isBlank(statusMessage)) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status message: &lt;" + (statusMessage != null ? htmlKit.escape(statusMessage) : "") + "&gt; is missing or is a blank expectedValue.");
                }
                break;

            case DoesNotExist:
                if (isBlank(statusMessage)) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status message: &lt;" + (htmlKit.escape(statusMessage)) + "&gt; exits.");
                }
                break;

            case Contains:
                if (!isBlank(statusMessage) && !isBlank(expectedValue) && statusMessage.toLowerCase().contains(expectedValue.toLowerCase())) {
                    assertion.setState(State.Ok);

                } else if (!isBlank(expectedValue)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status message: &lt;" + (statusMessage != null ? htmlKit.escape(statusMessage) : "") + "&gt; does not contain: &lt;" + htmlKit.escape(expectedValue) + "&gt;.");
                }
                break;

            case DoesNotContain:
                if (!isBlank(statusMessage) && !isBlank(expectedValue) && !statusMessage.toLowerCase().contains(expectedValue.toLowerCase())) {
                    assertion.setState(State.Ok);

                } else if (!isBlank(expectedValue) && !isBlank(statusMessage)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status message: &lt;" + (htmlKit.escape(statusMessage)) + "&gt; contains: &lt;" + htmlKit.escape(expectedValue) + "&gt;.");
                }
                break;

            default:
                break;
        }
        return Promises.of();
    }

    private Promise<Void> checkStatusCode(final AssertionResultTo assertion, final HttpResponseTo response) {

        Integer expectedValue = null;

        final String expectedValueAsString = !isBlank(assertion.getEvaluatedValue()) ? assertion.getEvaluatedValue() : assertion.getValue();

        if (!isBlank(expectedValueAsString)) {
            try {
                expectedValue = Integer.valueOf(expectedValueAsString.trim());

            } catch (final Exception e) {
                assertion.setState(State.Error);
                assertion.setActualValue(expectedValueAsString);
                assertion.addErrorMessage("Value &lt;" + htmlKit.escape(expectedValueAsString.trim()) + "&gt; is not a number.");
                return Promises.of();
            }
        }

        if (expectedValue == null) {
            return fillNotANumberInAssertion(assertion, expectedValueAsString);
        }

        if (response == null) {
            return Promises.of();
        }

        final Integer statusCode = response.getStatus() != null ? response.getStatus().getCode() : null;
        assertion.setActualValue(String.valueOf(statusCode));

        switch (assertion.getComparison()) {
            case Equals:
                if (statusCode == null) {
                    break;

                } else if (expectedValue.intValue() == statusCode.intValue()) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected response status code: &lt;" + expectedValue + "&gt; but was: &lt;" + statusCode + "&gt;.");
                }
                break;

            case DoesNotEqual:
                if (statusCode == null) {
                    break;

                } else if (expectedValue.intValue() != statusCode.intValue()) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status code: &lt;" + expectedValue + "&gt; is the same.");
                }
                break;

            case Less:
                if (statusCode == null) {
                    break;

                } else if (statusCode < expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status code: &lt;" + statusCode + "&gt; is not less than: &lt;" + expectedValue + "&gt;.");
                }
                break;

            case LessOrEqual:
                if (statusCode == null) {
                    break;

                } else if (statusCode <= expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status code: &lt;" + statusCode + "&gt; is not less or equal to: &lt;" + expectedValue + "&gt;.");
                }
                break;

            case Greater:
                if (statusCode == null) {
                    break;

                } else if (statusCode > expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status code: &lt;" + statusCode + "&gt; is not greater than: &lt;" + expectedValue + "&gt;.");
                }
                break;

            case GreateOrEqual:
                if (statusCode == null) {
                    break;

                } else if (statusCode >= expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response status code: &lt;" + statusCode + "&gt; is not greater or equal to: &lt;" + expectedValue + "&gt;.");
                }
                break;

            default:
                break;
        }
        return Promises.of();
    }

    private Promise<Void> fillNotANumberInAssertion(final AssertionResultTo assertion, final String expectedValueAsString) {
        assertion.setState(State.Error);
        assertion.setActualValue(expectedValueAsString);
        assertion.addErrorMessage("Value &lt;&gt; is not a number.");
        return Promises.of();
    }

    private Promise<Void> checkBodyLength(final AssertionResultTo assertion, final HttpResponseTo response) {

        Integer expectedValue = null;

        final String expectedValueAsString = !isBlank(assertion.getEvaluatedValue()) ? assertion.getEvaluatedValue() : assertion.getValue();

        if (!isBlank(expectedValueAsString)) {
            try {
                expectedValue = Integer.valueOf(expectedValueAsString.trim());

            } catch (final Exception e) {
                assertion.setState(State.Error);
                assertion.setActualValue(expectedValueAsString);
                assertion.addErrorMessage("Value &lt;" + htmlKit.escape(expectedValueAsString.trim()) + "&gt; is not a number.");
                return Promises.of();
            }
        }
        if (expectedValue == null) {
            return fillNotANumberInAssertion(assertion, expectedValueAsString);
        }

        if (response == null) {
            return Promises.of();
        }

        final int value = expectedValue;

        final Blob responseBlob = (response.getBody() == null || response.getBody().getBlob() == null) ? emptyBlob : response.getBody().getBlob();
        return checkBodyLength(assertion, value, responseBlob.size())
                .recoverOnReject();
    }

    private Promise<Void> checkBodyLength(final AssertionResultTo assertion, final int expectedValue, final int actualValue) {
        assertion.setActualValue(String.valueOf(actualValue));

        switch (assertion.getComparison()) {
            case Equals:
                if (actualValue == expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected response body length: &lt;" + expectedValue + "&gt; bytes but was: &lt;" + actualValue + "&gt; bytes.");
                }
                break;

            case DoesNotEqual:
                if (actualValue != expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body length: &lt;" + expectedValue + "&gt; bytes is the same.");
                }
                break;

            case Less:
                if (actualValue < expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body length: &lt;" + actualValue + "&gt; bytes is not less than: &lt;" + expectedValue + "&gt; bytes.");
                }
                break;

            case LessOrEqual:
                if (actualValue <= expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body length: &lt;" + actualValue + "&gt; bytes is not less or equal to: &lt;" + expectedValue + "&gt; bytes.");
                }
                break;

            case Greater:
                if (actualValue > expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body length: &lt;" + actualValue + "&gt; bytes is not greater than: &lt;" + expectedValue + "&gt; bytes.");
                }
                break;

            case GreateOrEqual:
                if (actualValue >= expectedValue) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body length: &lt;" + actualValue + "&gt; bytes is not greater or equal to: &lt;" + expectedValue + "&gt; bytes.");
                }
                break;

            default:
                break;
        }
        return Promises.of();
    }

    private Promise<Void> checkBodyContent(final AssertionResultTo assertion, final HttpResponseTo response) {

        if (response == null) {
            return Promises.of();
        }

        final String value = getNonBlankValue(assertion);

        final Blob responseBlob = (response.getBody() == null || response.getBody().getBlob() == null) ? emptyBlob : response.getBody().getBlob();

        return checkBodyContent(assertion, value, responseBlob)
                .recoverOnReject();
    }

    private String getNonBlankValue(final AssertionResultTo assertion) {
        if (!isBlank(assertion.getEvaluatedValue())) {
            return assertion.getEvaluatedValue().trim();
        }

        return !isBlank(assertion.getValue()) ? assertion.getValue().trim() : null;
    }

    private Promise<Void> checkBodyContent(final AssertionResultTo assertion, final String expectedValue, final Blob actualContentBlob) {
        if (assertion == null) {
            return Promises.of();
        }

        return BlobUtils.read(blobReader, actualContentBlob)
                .doOnReject(new Consumer<Throwable>() {
                    @Override
                    public void consume(final Throwable throwable) {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Cannot read response body.");
                    }
                })
                .doOnResolve(new Consumer<String>() {
                    @Override
                    public void consume(final String actualContent) {
                        checkBodyContent(assertion, expectedValue, actualContent);
                    }
                })
                .castToVoidPromise();
    }

    private void checkBodyContent(final AssertionResultTo assertion, final String expectedValue, final String actualContent) {
        assertion.setActualValue(actualContent);
        assertion.setState(State.Ok);
        switch (assertion.getComparison()) {
            case Equals:
                if (isBlank(actualContent)) {
                    if (!isBlank(expectedValue)) {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Expected response body &lt;" + (htmlKit.escape(expectedValue)) + "&gt; but was empty.");
                    }
                } else if (isBlank(expectedValue)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected empty body but body length was: &lt;" + actualContent.length() + "&gt; bytes.");
                } else if (compare(actualContent, expectedValue.trim()) != 0) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected response body &lt;" + (htmlKit.escape(expectedValue.trim())) + "&gt; does not match.");
                }
                break;
            case DoesNotEqual:
                if (isBlank(expectedValue) && isBlank(actualContent)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body is the same.");
                } else if (!isBlank(actualContent)
                        && !isBlank(expectedValue)
                        && compare(actualContent, expectedValue.trim()) == 0) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected response body &lt;" + (htmlKit.escape(expectedValue.trim())) + "&gt; does not match.");
                }
                break;
            case Exists:
                if (isBlank(actualContent)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body does not exist.");
                }
                break;

            case DoesNotExist:
                if (!isBlank(actualContent)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body exists.");
                }
                break;

            case Contains:
                if (!isBlank(expectedValue)) {
                    if ((!isBlank(actualContent) && !actualContent.contains(expectedValue.trim())
                            || isBlank(actualContent))) {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Response body does not contain &lt;" + (htmlKit.escape(expectedValue.trim())) + "&gt;.");
                    }
                }
                break;

            case DoesNotContain:
                if (!isBlank(expectedValue)
                        && !isBlank(actualContent)
                        && actualContent.contains(expectedValue.trim())) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body contains &lt;" + (htmlKit.escape(expectedValue.trim())) + "&gt;.");
                }
                break;
            default:
                break;
        }
    }

    private Promise<Void> checkHeader(final AssertionResultTo assertion, final HttpResponseTo response) {
        if (response == null) {
            return Promises.of();
        }

        final String headerName = firstNotBlank(assertion.getEvaluatedPath(), assertion.getPath()).orElse(null);

        if (isBlank(headerName)) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("Response header name is required.");
            return Promises.of();
        }

        final String expectedHeaderValue = !isBlank(assertion.getEvaluatedValue()) ? assertion.getEvaluatedValue() : assertion.getValue();

        final List<String> headerValues = findHeaderValue(response.getHeaders(), headerName);
        final String headerValue = isNullOrEmpty(headerValues) ? null : joiner(headerValues).joinWith(",");

        assertion.setActualValue(headerValue);

        switch (assertion.getComparison()) {
            case Equals:
                if (headerValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");

                } else if ((!isBlank(expectedHeaderValue) && !isBlank(headerValue) && compare(expectedHeaderValue, headerValue) == 0)) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected header &lt;" + htmlKit.escape(headerName) + "&gt; value: &lt;" + (expectedHeaderValue != null ? htmlKit.escape(expectedHeaderValue) : "") + "&gt; but was: &lt;" + (htmlKit.escape(headerValue)) + "&gt;.");
                }
                break;

            case DoesNotEqual:
                if (headerValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");

                } else if ((!isBlank(expectedHeaderValue) && !isBlank(headerValue) && compare(expectedHeaderValue, headerValue) == 0)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; value: &lt;" + (htmlKit.escape(headerValue)) + "&gt; is the same.");

                } else {
                    assertion.setState(State.Ok);
                }
                break;

            case Exists:
                if (headerValue != null) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");
                }
                break;

            case DoesNotExist:
                if (headerValue == null) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; exists.");
                }
                break;

            case Contains:
                if (headerValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");

                } else if (!isBlank(headerValue) && !isBlank(expectedHeaderValue) && headerValue.toLowerCase().contains(expectedHeaderValue.toLowerCase())) {
                    assertion.setState(State.Ok);

                } else if (!isBlank(expectedHeaderValue)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; value: &lt;" + (htmlKit.escape(headerValue)) + "&gt; does not contain: &lt;" + (htmlKit.escape(expectedHeaderValue)) + "&gt;.");
                }
                break;

            case DoesNotContain:
                if (headerValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");

                } else if (!isBlank(headerValue) && !isBlank(expectedHeaderValue) && !headerValue.toLowerCase().contains(expectedHeaderValue.toLowerCase())) {
                    assertion.setState(State.Ok);

                } else if (!isBlank(expectedHeaderValue)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; value: &lt;" + (htmlKit.escape(headerValue)) + "&gt; contains: &lt;" + (htmlKit.escape(expectedHeaderValue)) + "&gt;.");
                }
                break;

            case Less:
                if (headerValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");

                } else if (!isBlank(headerValue) && !isBlank(expectedHeaderValue) && compare(headerValue, expectedHeaderValue) < 0) {
                    assertion.setState(State.Ok);

                } else if (!isBlank(expectedHeaderValue)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; value: &lt;" + (htmlKit.escape(headerValue)) + "&gt; is not less than: &lt;" + (htmlKit.escape(expectedHeaderValue)) + "&gt;.");
                }
                break;

            case LessOrEqual:
                if (headerValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");

                } else if (!isBlank(headerValue) && !isBlank(expectedHeaderValue) && compare(headerValue, expectedHeaderValue) <= 0) {
                    assertion.setState(State.Ok);

                } else {
                    if (!isBlank(expectedHeaderValue)) {
                        assertion.setState(State.Failure);
                        assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; value: &lt;" + (htmlKit.escape(headerValue)) + "&gt; is not less or equal to: &lt;" + (htmlKit.escape(expectedHeaderValue)) + "&gt;.");
                    }
                }
                break;

            case Greater:
                if (headerValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");

                } else if (!isBlank(headerValue) && !isBlank(expectedHeaderValue) && compare(headerValue, expectedHeaderValue) > 0) {
                    assertion.setState(State.Ok);

                } else if (!isBlank(expectedHeaderValue)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; value: &lt;" + (htmlKit.escape(headerValue)) + "&gt; is not greater than: &lt;" + (htmlKit.escape(expectedHeaderValue)) + "&gt;.");
                }
                break;

            case GreateOrEqual:
                if (headerValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; does not exist.");

                } else if (!isBlank(headerValue) && !isBlank(expectedHeaderValue) && compare(headerValue, expectedHeaderValue) >= 0) {
                    assertion.setState(State.Ok);

                } else if (!isBlank(expectedHeaderValue)) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response header &lt;" + htmlKit.escape(headerName) + "&gt; value: &lt;" + (htmlKit.escape(headerValue)) + "&gt; is not greater or equal to: &lt;" + (htmlKit.escape(expectedHeaderValue)) + "&gt;.");
                }
                break;

            default:
                break;
        }
        return Promises.of();
    }

    private int compare(final String v1, final String v2) {
        if (!isBlank(v1) && !isBlank(v2)) {

            Maybe<Integer> result = Maybe.ofThrowable(new BiFunction<String, String, Integer>() {
                @Override
                public Integer apply(final String firstParam, final String secondParam) {
                    final Long l1 = Long.valueOf(firstParam);
                    final Long l2 = Long.valueOf(secondParam);
                    return l1.compareTo(l2);
                }
            }, v1, v2);

            if (result.isPresent()) {
                return result.get();
            }

            result = Maybe.ofThrowable(new BiFunction<String, String, Integer>() {
                @Override
                public Integer apply(final String firstParam, final String secondParam) {
                    final Double d1 = Double.valueOf(firstParam);
                    final Double d2 = Double.valueOf(secondParam);
                    return d1.compareTo(d2);
                }
            }, v1, v2);

            if (result.isPresent()) {
                return result.get();
            }

            final Date d1 = httpDate.valueOf(v1);
            final Date d2 = httpDate.valueOf(v2);

            if (d1 != null && d2 != null) {
                return d1.compareTo(d2);
            }

            return v1.compareToIgnoreCase(v2);

        } else if (isBlank(v1) && isBlank(v2)) {
            return 0;

        } else if (!isBlank(v1)) {
            return 1;
        }

        return -1;
    }

    private static final Blob emptyBlob = new Blob() {
        @Override
        public int size() {
            return 0;
        }

        @Override
        public String type() {
            return null;
        }

        @Override
        public String name() {
            return null;
        }

        @Override
        public boolean isInferredType() {
            return false;
        }

        @Override
        public boolean isEmpty() {
            return true;
        }
    };

}
