package com.restlet.client.tests.asserts.impl;

import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.async.impl.DeferredImpl;
import com.restlet.client.html.HtmlKit;
import com.restlet.client.net.http.HttpDate;
import com.restlet.client.net.http.response.HttpResponseTo;
import com.restlet.client.platform.blob.BlobReader;
import com.restlet.client.platform.xml.XPathResult;
import com.restlet.client.platform.xml.XPathResult.XPathResultType;
import com.restlet.client.platform.xml.XmlEngine;
import com.restlet.client.platform.xml.XmlNode;
import com.restlet.client.tests.asserts.AssertionResultTo;
import com.restlet.client.tests.asserts.AssertionResultTo.State;
import com.restlet.client.tests.asserts.ResponseBodyAssertionEngine;
import com.restlet.client.utils.Maybe;

import java.util.Collection;
import java.util.Date;

import static com.restlet.client.utils.StringUtils.firstNotBlank;
import static com.restlet.client.utils.StringUtils.isBlank;

public class XmlAssertionEngineImpl implements ResponseBodyAssertionEngine {

    private final XmlEngine xmlEngine;
    private final HttpDate httpDate;
    private final BlobReader blobReader;
    private final HtmlKit htmlKit;

    public XmlAssertionEngineImpl(final XmlEngine xmlEngine, final HttpDate httpDate, final BlobReader blobReader, final HtmlKit htmlKit) {
        super();
        this.xmlEngine = xmlEngine;
        this.httpDate = httpDate;
        this.blobReader = blobReader;
        this.htmlKit = htmlKit;
    }

    @Override
    public Promise<Void> checkBody(final AssertionResultTo assertion, final HttpResponseTo response) {
        if (response == null) {
            return Promises.of();
        }

        if (isBlank(assertion.getPath())) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> is required.");
            return Promises.of();
        }

        if (response.getBody() == null || response.getBody().getBlob() == null || response.getBody().getBlob().size() <= 0) {
            assertion.setState(State.Failure);
            assertion.addFailureMessage("Response body does not exist.");
            return Promises.of();
        }

        final Deferred<Void> deferred = new DeferredImpl<>();

        blobReader.readAsText(response.getBody().getBlob(), new BlobReader.Handler<String>() {

            @Override
            public void onLoaded(final String responseBodyAsText) {
                if (!isBlank(responseBodyAsText)) {
                    assertion.setActualValue(responseBodyAsText);
                    matchXml(responseBodyAsText, assertion);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body is not XML.");
                }
                deferred.resolve();
            }

            @Override
            public void onError(final Object o) {
                deferred.resolve();

            }
        });
        return deferred.promise();
    }

    private boolean matchXml(final String xml, final AssertionResultTo assertion) {
        final String stringPath = firstNotBlank(assertion.getEvaluatedPath(), assertion.getPath()).orElse(null);

        if (isBlank(stringPath)) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> is required.");
            return false;
        }

        //TODO pre-validate XPath syntax

        final XmlNode node;
        try {
            node = Maybe.of(xmlEngine.fromXml(xml)).get();
        } catch (final Throwable e) {
            assertion.setState(State.Failure);
            assertion.addFailureMessage("Response body is not valid XML.");
            return false;
        }

        final XPathResult xPathResult;
        try {
            xPathResult = Maybe.of(xmlEngine.xpath(node, stringPath)).get();
        } catch (final Throwable e) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("'" + stringPath + "' is not valid <a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a>");
            return false;
        }

        switch (assertion.getComparison()) {
            case Exists:
                if (!xPathResult.isEmpty()) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target does not exist or is a null.");
                }
                return State.Ok == assertion.getState();

            case DoesNotExist:
                if (xPathResult.isEmpty()) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target exists, &lt;" + htmlEscape(xPathResult, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            default:
                break;
        }

        if (xPathResult.getResultType() == null) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("Unknown XPath result type.");
            return false;
        }

        final String stringValue = firstNotBlank(assertion.getEvaluatedValue(), assertion.getValue()).orElse(null);

        if (xPathResult.getResultType() == XPathResultType.String
                || xPathResult.getResultType() == XPathResultType.Boolean
                || xPathResult.getResultType() == XPathResultType.Number) {

            Object expectedValue = stringValue;
            Object sourceValue = null;

            switch (xPathResult.getResultType()) {
                case String:
                    sourceValue = xPathResult.getStringValue();
                    break;

                case Number:
                    sourceValue = xPathResult.getNumberValue();
                    try {
                        expectedValue = Double.valueOf(stringValue);
                    } catch (final Exception e) {
                        // do nothing, checks will be done with string value
                    }
                    break;

                case Boolean:
                    sourceValue = xPathResult.getBooleanValue();
                    try {
                        expectedValue = Boolean.valueOf(stringValue);
                    } catch (final Exception e) {
                        // do nothing, checks will be done with string value
                    }
                    break;

                default:
                    break;
            }

            final Boolean validated = validate(assertion, sourceValue, expectedValue);
            if (validated != null) {
                return validated;
            }
        }

        Boolean passed = null;

        final Collection<XmlNode> sourceNodes = xPathResult.getMatchingNodes();

        if (sourceNodes != null) {

            for (final XmlNode n : sourceNodes) {

                final Boolean p = validate(assertion, n.toString().trim().replaceAll(">\\s*", ">").replaceAll("\\s<", ""), stringValue);

                if (p != null) {
                    if (passed == null) {
                        passed = p;
                    } else {
                        passed = passed && p;
                    }
                }

                if (passed != null && !passed) {
                    return false;
                }
            }
        }

        if (passed == null) {
            assertion.setState(State.Failure);
            assertion.addFailureMessage("Nothing found for the given XPath. [" + stringPath + "]");
        }

        return false;
    }

    public Boolean validate(final AssertionResultTo assertion, final Object sourceValue, final Object expectedValue) {
        switch (assertion.getComparison()) {
            case Exists:
                if (sourceValue != null) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target does not exist or is a null.");
                }
                return State.Ok == assertion.getState();

            case DoesNotExist:
                if (sourceValue == null) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target exists, &lt;" + htmlEscape(sourceValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case Equals:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (equals(sourceValue, expectedValue)) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected sourceValue: &lt;" + htmlEscape(expectedValue, "null") + "&gt; but was: &lt;" + htmlEscape(sourceValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case DoesNotEqual:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (!equals(sourceValue, expectedValue)) {
                    assertion.setState(State.Ok);

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value: &lt;" + htmlEscape(sourceValue, "null") + "&gt is the same;.");
                }
                return State.Ok == assertion.getState();

            case Contains:
                if (sourceValue != null && expectedValue != null && contains(sourceValue, expectedValue)) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; does not contain &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case DoesNotContain:
                if (sourceValue != null && expectedValue != null && !contains(sourceValue, expectedValue)) {
                    assertion.setState(State.Ok);
                    return true;

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target does not exist or is a null.");

                } else if (expectedValue != null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; contains &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case Less:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (sourceValue != null && expectedValue != null && compare(sourceValue, expectedValue) < 0) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; is not less than &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case LessOrEqual:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (sourceValue != null && expectedValue != null && compare(sourceValue, expectedValue) <= 0) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; is not less or equal to &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case Greater:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (sourceValue != null && expectedValue != null && compare(sourceValue, expectedValue) > 0) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; is not greater than &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case GreateOrEqual:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (sourceValue != null && expectedValue != null && compare(sourceValue, expectedValue) >= 0) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"https://www.w3.org/TR/xpath/\" target=\"_blank\">XPath 1.0</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; is not greater or equal to &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            default:
                break;
        }
        return null;
    }

    private void warnUserTypesDiffer(final AssertionResultTo assertion, final Object value, final Object expected) {
        if (expected == null || value == null) {
            return;
        }
        if (!haveSameType(value, expected)) {
            assertion.addWarningMessage("Assertion compares values as strings (<a href=\"https://en.wikipedia.org/wiki/Lexicographical_order\" target=\"_blank\">lexicographically</a>) since &lt;" + htmlEscape(value) + "&gt; and &lt;" + htmlEscape(expected, "null") + "&gt; have different types (&lt;" + htmlEscape(value) + "&gt; has type \"" + getType(value) + "\" and &lt;" + htmlEscape(expected, "null") + "&gt; has type \"" + getType(expected) + "\")");
        }
    }

    private boolean haveSameType(final Object v1, final Object v2) {
        return areBooleanValues(v1, v2)
                || areDoubleValues(v1, v2)
                || areStringValues(v1, v2);
    }

    private String getType(final Object value) {
        if (value instanceof Boolean) {
            return "boolean";
        } else if (value instanceof Double) {
            return "number";
        } else if (value instanceof String) {
            return "string";
        }
        return "object";
    }

    private boolean equals(final Object v1, final Object v2) {
        if (v1 == null && v2 == null) {
            return true;
        } else if (v1 == null) {
            return false;
        } else if (v2 == null) {
            return false;
        }

        if (areBooleanValues(v1, v2)) {
            return ((Boolean) v1).booleanValue() == ((Boolean) v2).booleanValue();
        }

        if (areDoubleValues(v1, v2)) {
            return ((Double) v1).doubleValue() == ((Double) v2).doubleValue();
        }

        if (areStringValues(v1, v2)) {
            final String s1 = v1.toString();
            final String s2 = v2.toString();

            final Date d1 = httpDate.valueOf(s1);
            final Date d2 = httpDate.valueOf(s2);

            if (d1 != null && d2 != null) {
                return d1.equals(d2);
            }
            return s1.equalsIgnoreCase(s2);
        }

        final String s1 = v1.toString();
        final String s2 = v2.toString();
        return (s1 != null && s2 != null) && s1.equalsIgnoreCase(s2);
    }

    public int compare(final Object v1, final Object v2) {
        if (v1 == null && v2 == null) {
            return 0;
        } else if (v1 == null) {
            return -1;
        } else if (v2 == null) {
            return 1;
        }

        if (areBooleanValues(v1, v2)) {

            if ((((Boolean) v1) && ((Boolean) v2) || !((Boolean) v1) && !((Boolean) v2))) {
                return 0;
            }
            if ((Boolean) v1) {
                return 1;
            }
            return -1;
        }

        if (areDoubleValues(v1, v2)) {
            return ((Double) v1).compareTo((Double) v2);
        }

        if (areStringValues(v1, v2)) {
            try {
                final Long l1 = Long.valueOf(v1.toString());
                final Long l2 = Long.valueOf(v2.toString());

                if (l1 != null && l2 != null) {
                    return l1.compareTo(l2);
                }

            } catch (final Exception e) {
                // do nothing, let's try to convert to another type
            }

            try {
                final Double d1 = Double.valueOf(v1.toString());
                final Double d2 = Double.valueOf(v2.toString());

                if (d1 != null && d2 != null) {
                    return d1.compareTo(d2);
                }
            } catch (final Exception e) {
                // do nothing, let's try to convert to another type
            }

            final Date d1 = httpDate.valueOf(v1.toString());
            final Date d2 = httpDate.valueOf(v2.toString());

            if (d1 != null && d2 != null) {
                return d1.compareTo(d2);
            }

            return v1.toString().compareToIgnoreCase(v2.toString());
        }

        final String s1 = v1.toString();
        final String s2 = v2.toString();

        return (s1 != null && s2 != null) ? s1.compareToIgnoreCase(s2) : (s1 == null ? -1 : 1);
    }

    private boolean areBooleanValues(final Object v1, final Object v2) {
        return v1 instanceof Boolean && v2 instanceof Boolean;
    }

    private boolean areDoubleValues(final Object v1, final Object v2) {
        return v1 instanceof Double && v2 instanceof Double;
    }

    private boolean areStringValues(final Object v1, final Object v2) {
        return v1 instanceof String && v2 instanceof String;
    }

    private boolean contains(final Object o1, final Object o2) {
        if (o1 != null && o2 != null) {

            final String s1 = (o1 instanceof String) ? (String) o1 : o1.toString();
            final String s2 = (o2 instanceof String) ? (String) o2 : o2.toString();

            if (s1 != null && s2 != null) {
                return s1.toLowerCase().contains(s2.toLowerCase());
            }
        }
        return false;
    }

    private String htmlEscape(final Object value, final String defaultValue) {
        return value == null ? defaultValue : htmlKit.escape(value.toString());
    }

    private String htmlEscape(final Object value) {
        return htmlEscape(value, null);
    }

}
