package com.restlet.client.utils;

import com.restlet.client.function.Consumer;
import com.restlet.client.function.Predicate;
import com.restlet.client.html.form.FormItemTo;
import com.restlet.client.html.form.FormItemType;
import com.restlet.client.html.form.FormTo;
import com.restlet.client.model.EntityTo;
import com.restlet.client.model.EntityTreeNode;
import com.restlet.client.model.environment.EnvironmentTo;
import com.restlet.client.model.environment.EnvironmentVariableTo;
import com.restlet.client.tests.ScenarioTo;
import com.restlet.client.tests.asserts.AssertionResultTo;
import com.restlet.client.tests.asserts.AssertionTo;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import static com.restlet.client.model.EntityTo.Type.getEntityTypeComparator;
import static com.restlet.client.model.UpdateType.None;
import static com.restlet.client.net.http.HeaderUtils.MEDIA_TYPE_MULTIPART_FORM_DATA;
import static com.restlet.client.net.http.HeaderUtils.MEDIA_TYPE_WEB_FORM;

public class EntityUtils {

    public static Maybe<EntityTo> getParent(final EntityTo entity, final EntityTo.Type type) {
        EntityTo result = entity;
        while (result != null) {
            if (result.getType() == type) {
                return Maybe.of(result);
            }
            result = result.getParent();
        }
        return Maybe.ofNullable(null);
    }

    public static EntityTreeNode recursiveFindParents(final EntityTo parent) {
        return recursiveFindParents(null, parent);
    }

    public static EntityTreeNode recursiveFindParents(final EntityTreeNode preparedTreeNode, final EntityTo parent) {
        if (parent == null) {
            return preparedTreeNode;
        }

        final EntityTreeNode wrappedTreeNode = new EntityTreeNode(parent, None);
        wrappedTreeNode.children.add(preparedTreeNode);
        return recursiveFindParents(wrappedTreeNode, parent.getParent());
    }

    public static EntityTo getRootEntity(final EntityTo entity) {
        EntityTo result = entity;
        while (result.getParent() != null) {
            result = result.getParent();
        }
        return result;
    }

    private static boolean contains(final EnvironmentTo environment, final String variableName) {
        return Sequence.of(environment.getVariables())
                .some(new Predicate<EnvironmentVariableTo>() {
                    @Override
                    public boolean test(final EnvironmentVariableTo environmentVariable) {
                        return Objects.equals(variableName, environmentVariable.getName());
                    }
                });
    }

    public static boolean addEnvironmentVariable(final EnvironmentTo environment, final EnvironmentVariableTo variable) {
        List<EnvironmentVariableTo> variables = environment.getVariables();
        if (variables == null) {
            variables = new ArrayList<>();
            environment.setVariables(variables);
        }
        if (contains(environment, variable.getName())) {
            return false;
        }
        variables.add(variable);
        return true;
    }

    public static int indexOf(final ScenarioTo scenarioTo, final EntityTo request) {
        return (request == null) ? -1 : indexOf(scenarioTo, request.getName());
    }

    public static int indexOf(final ScenarioTo scenarioTo, final String requestName) {
        return (scenarioTo.getRequestOrder() == null) ? -1 : scenarioTo.getRequestOrder().indexOf(requestName);
    }

    public static final Comparator<EntityTo> ENTITY_COMPARATOR = new Comparator<EntityTo>() {
        @Override
        public int compare(final EntityTo entity1, final EntityTo entity2) {
            final int typeComparison = -getEntityTypeComparator().compare(entity1.getType(), entity2.getType());

            return typeComparison == 0
                    ? entity1.getName().compareTo(entity2.getName())
                    : typeComparison;
        }
    };

    public static final Comparator<EntityTreeNode> ENTITY_TREE_NODE_COMPARATOR = new Comparator<EntityTreeNode>() {
        @Override
        public int compare(final EntityTreeNode entityTreeNode1, final EntityTreeNode entityTreeNode2) {
            return ENTITY_COMPARATOR.compare(entityTreeNode1.entity, entityTreeNode2.entity);
        }
    };

    public static boolean equalById(final EntityTo entity1, final EntityTo entity2) {
        if (entity1 == null) {
            return entity2 == null;

        } else if (entity2 == null) {
            return false;
        }

        return Objects.equals(entity1.getId(), entity2.getId());
    }

    public static boolean isEnabled(final AssertionTo assertion) {
        return assertion.isEnabled() == null || assertion.isEnabled();
    }

    public static boolean isEnabled(final AssertionResultTo assertionResult) {
        return assertionResult.isEnabled() == null || assertionResult.isEnabled();
    }

    public static String generateDottedName(@NotNull final EntityTo entity) {
        final StringBuilder dottedName = new StringBuilder();
        final String name = entity.getName();
        dottedName.append(name);
        EntityTo parent = entity.getParent();
        while (parent != null) {
            dottedName.insert(0, ".");
            dottedName.insert(0, parent.getName());
            parent = parent.getParent();
        }

        return dottedName.toString();
    }

    public static boolean isScenario(final EntityTo entityTo) {
        return entityTo instanceof ScenarioTo;
    }

    public static void walkParents(final EntityTo entityTo, final Consumer<EntityTo> visitor) {
        if (entityTo != null) {
            visitor.consume(entityTo);
            walkParents(entityTo.getParent(), visitor);
        }
    }

    public static String detectEncoding(final FormTo formTo) {
        return formTo.getEncoding() != null
                ? formTo.getEncoding().trim()
                : detectEncoding(formTo.getItems());
    }

    public static String detectEncoding(final List<FormItemTo> items) {
        if (Sequence.of(items).some(new Predicate<FormItemTo>() {
            @Override
            public boolean test(final FormItemTo formItemTo) {
                return FormItemType.File == formItemTo.getType();
            }
        })) {
            return MEDIA_TYPE_MULTIPART_FORM_DATA;
        }

        return MEDIA_TYPE_WEB_FORM;
    }

}
