package com.restlet.client.utils;

import com.restlet.client.function.Function;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import static java.util.Arrays.asList;

public class StringUtils {

    private static final String FORMAT_PLACEHOLDER = "%s";

    public static boolean isBlank(final String s) {
        return (s == null) || (s.trim().isEmpty());
    }

    public static boolean containsIgnoringCase(final String container, final String contained) {
        if (container == null) {
            return false;
        }

        if (isBlank(contained)) {
            return true;
        }

        return container.toLowerCase().contains(contained.toLowerCase());
    }

    /**
     * Calculates precise bytes in string. Takes around 200ms per 1mb string
     * but is much faster and memory efficient then String.getBytes().length
     *
     * @param str string
     * @return number of bytes in string
     */
    public static int getUTF8StringByteCount(@NotNull final String str) {
        //code below is from gwt supersource for String.getBytesUTF8
        final int n = str.length();
        int byteCount = 0;
        for (int i = 0; i < n; ) {
            final int ch = str.codePointAt(i);
            i += Character.charCount(ch);
            if (ch < (1 << 7)) {
                byteCount++;
            } else if (ch < (1 << 11)) {
                byteCount += 2;
            } else if (ch < (1 << 16)) {
                byteCount += 3;
            } else if (ch < (1 << 21)) {
                byteCount += 4;
            } else if (ch < (1 << 26)) {
                byteCount += 5;
            }
        }
        return byteCount;
    }

    public static String truncate(final String description, final int i) {
        if (StringUtils.isBlank(description)) {
            return "";
        }
        if (description.length() > i) {
            return description.substring(0, i - 4).concat("\n...");
        }
        return description;
    }

    public static String firstUpper(final String string) {
        if (!isBlank(string)) {
            return string.substring(0, 1).toUpperCase() + string.substring(1);
        }
        return string;
    }

    public static String trim(final String s) {
        return s == null ? null : s.trim();
    }

    public static boolean oneNotBlank(final String... values) {
        for (final String value : values) {
            if (!isBlank(value)) {
                return true;
            }
        }
        return false;
    }

    public static Maybe<String> firstNotBlank(final String... strings) {
        for (final String s : strings) {
            if (!isBlank(s)) {
                return Maybe.of(s.trim());
            }
        }
        return Maybe.empty();
    }

    public static String firstNotNull(final String... strings) {
        for (final String string : strings) {
            if (string != null) {
                return string;
            }
        }
        throw new IllegalArgumentException("One input at least should be non-null");
    }

    public static String trimLeft(final String string, final char charToTrim) {
        if (isBlank(string)) {
            return string;
        }
        return string.replaceFirst("^" + charToTrim + "+", "");
    }

    private static <T> Function<T, String> toStringFunction() {
        return new Function<T, String>() {
            @Override
            public String apply(final T input) {
                if (input == null) {
                    return null;
                }

                return input instanceof String
                        ? (String) input
                        : input.toString();
            }
        };
    }

    public static <T> Joiner joiner(final List<T> parts) {
        return new Joiner(new AsStringIterator<>(parts.iterator(), StringUtils.<T>toStringFunction()));
    }

    @SafeVarargs
    public static <T> Joiner joiner(final T... parts) {
        return new Joiner(new AsStringIterator<>(asList(parts).iterator(), StringUtils.<T>toStringFunction()));
    }

    public static <T> Joiner joiner(final Iterable<T> parts, final Function<T, String> toString) {
        return new Joiner(new AsStringIterator<>(parts.iterator(), toString));
    }

    public static String padLeft(String value, final int length, final char c) {
        if (value == null) {
            value = "";
        }
        if (value.length() >= length) {
            return value;
        }
        final char[] tab = new char[length - value.length()];
        Arrays.fill(tab, c);
        return new String(tab) + value;
    }

    public static String format(final String source, final Object... replacements) {
        Objects.requireNonNull(replacements);

        String result = source;

        for (final Object replacement : replacements) {
            final int indexOfPlaceholder = result.indexOf(FORMAT_PLACEHOLDER);
            if (indexOfPlaceholder == -1) {
                throw new IllegalArgumentException("Missing placeholder");
            }
            result = result.substring(0, indexOfPlaceholder) + replacement.toString() + result.substring(indexOfPlaceholder + FORMAT_PLACEHOLDER.length());
        }

        if (result.contains(FORMAT_PLACEHOLDER)) {
            throw new IllegalArgumentException("At least one placeholder had no value, current state is " + result);
        }

        return result;
    }

    public static class Joiner {

        private final Iterator<String> iterator;

        private Joiner(final Iterator<String> iterator) {
            this.iterator = iterator;
        }

        public String joinWith(final String separator) {
            final StringBuilder joined = new StringBuilder();

            if (iterator.hasNext()) {
                joined.append(iterator.next());
            }

            while (iterator.hasNext()) {
                joined
                        .append(separator)
                        .append(iterator.next());
            }

            return joined.toString();
        }
    }

    private static class AsStringIterator<T> implements Iterator<String> {

        private final Iterator<T> wrappedIterator;
        private final Function<T, String> toString;

        AsStringIterator(final Iterator<T> wrappedIterator, final Function<T, String> toString) {
            this.wrappedIterator = wrappedIterator;
            this.toString = toString;
        }

        @Override
        public boolean hasNext() {
            return wrappedIterator.hasNext();
        }

        @Override
        public String next() {
            return toString.apply(wrappedIterator.next());
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("remove");
        }
    }
}
